@@ -719,7 +719,7 @@ static int ibmca_signature_rsa_sign(void *vctx,
         goto out;
 
     /* Perform private key encrypt */
-    rc = ibmca_rsa_crt_with_blinding(ctx->key, enc_data, sig, rsa_size);
+    rc = ibmca_rsa_priv_with_blinding(ctx->key, enc_data, sig, rsa_size);
     if (rc != 1) {
         ibmca_debug_op_ctx(ctx, "ibmca_asym_cipher_rsa_with_blinding failed");
 
